"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const input_ui_ipc_provider_1 = require("input-ui-ipc-provider");
const ts_simple_ast_extra_1 = require("ts-simple-ast-extra");
const refactors_1 = require("../refactors");
const SignatureAbstractCodeFix_1 = require("../SignatureAbstractCodeFix");
const util_1 = require("../util");
/**
 * # Description
 *
 *  Allows users to change a signature parameter order. see README
 *
 * # TODO
 *  * refactor a implementation method wont change its interface signature - super
 *  * constructors not supported
    * we could give a more intuitive text-based API by letting the user provide the new signature. Then we create a new function with that signature in order to parse it correctly.
    * visual feedback select the arguments after refactor and put the cursor where in the middle
 */
class ReorderParamsCodeFixImpl extends SignatureAbstractCodeFix_1.SignatureAbstractCodeFix {
    constructor() {
        super(...arguments);
        this.name = refactors_1.PLUGIN_NAME + '-reorderParams';
        this.config = {
            // print a comment with help on the autocomplete suggestion
            help: false
        };
        this.helpText = `Help: The second argument is the new order of parameters. Number N in index M means move the M-th argument to index N. Examples: * [1] means switch between first and second  * [3, 2] means move the first parameter to fourth position and move the second parameter to the third. (the rest of the parameters, (third and fourth) will move left to accommodate this requirements)`;
    }
    getTargetInfo(sourceFile, position) {
        const targetInfoPredicate = (targetNode) => !(!targetNode || targetNode.parameters && targetNode.parameters.length <= 1) && !(!targetNode || targetNode.arguments && targetNode.arguments.length <= 1);
        return util_1.getTargetInfo(sourceFile, position, targetInfoPredicate);
    }
    printRefactorSuggestionMessage(targetInfo) {
        return `Reorder parameters of "${targetInfo ? targetInfo.name : this.targetInfo.name}"`;
    }
    getTextUIToolConfig() {
        return this.textUIToolConfigFactory({
            name: 'reorderParams',
            args: ['name', 'reorder'],
            snippet: (fileName, position) => {
                const sourceFile = this.options.info.languageService.getProgram().getSourceFile(fileName);
                this.targetInfo = this.getTargetInfo(sourceFile, position);
                if (!this.targetInfo) {
                    return;
                }
                const help = this.config.helpComment ? this.helpComment() : '';
                return `reorderParams("${this.targetInfo.name}", ${this.printReorderExample()})${help}`;
            },
        });
    }
    printReorderExample() {
        return '[' + new Array(this.targetInfo.parameterCount || 1).fill(0).map((v, i, arr) => arr.length - i - 1).join(', ') + ']';
    }
    helpComment() {
        return `
    
  /* ${this.helpText} */`;
    }
    apply(arg) {
        if (!this.selectedAction && this.inputConsumer.hasSupport(input_ui_ipc_provider_1.INPUT_ACTIONS.inputText)) {
            const defaultValue = this.printReorderExample();
            const options = {
                prompt: 'Enter reorderParam definition',
                placeHolder: defaultValue,
                value: defaultValue,
                description: this.helpText,
                validateInput: (function validateInput(s) { try {
                    JSON.parse(s);
                    return Promise.resolve();
                }
                catch (ex) {
                    return Promise.resolve('JSON is invalid, fix it please');
                } }).toString()
            };
            this.inputConsumer.inputText(options)
                .then(response => {
                const reorder = JSON.parse(response.answer);
                this.applyImpl(arg, (n) => { reorderParameters(n, reorder, this.options.log); });
            })
                .catch(ex => {
                this.options.log('this.inputConsumer.inputText catch ' + ex);
            });
        }
        else {
            this.applyImpl(arg, (n) => { reorderParameters(n, this.selectedAction.args.reorder, this.options.log); });
        }
    }
}
exports.ReorderParamsCodeFixImpl = ReorderParamsCodeFixImpl;
// the following code isolates the AST manipulation for this plugin apply()
/**
 * For all references that must be refactored calls reorderArgsOrParams.
 *
 * @param targetDeclaration the function-like declaration to reorder its parameters
 * @param reorder [1,0] means switching the positions between first and second params
 */
function reorderParameters(node, reorder, log) {
    log(`reorderParameters called with reorder: [${reorder.join(', ')}]`);
    util_1.applyToSignature(node, (argsOrParams, parent, returnValue) => reorderArgsOrParams(reorder, argsOrParams, log), log);
}
exports.reorderParameters = reorderParameters;
function reorderArgsOrParams(reorder, args, log) {
    // will use indexOccupied to reassign new positions to those nodes that must move implicitly
    let indexOccupied = new Array(args.length).fill(-1);
    reorder.forEach((r, i) => {
        if (r >= indexOccupied.length) {
            //user error provided incorrect reorder array
        }
        indexOccupied[r] = i;
    });
    function findNextFreeIndexFor(index) {
        const id = indexOccupied.indexOf(-1);
        indexOccupied[id] = index;
        return id;
    }
    // reorderedArgs will have metadata regarding the new params/args locations
    let reorderedArgs = args.map((arg, i) => {
        if (i < reorder.length) {
            return {
                index: reorder[i],
                originalIndex: i,
                arg: arg.getText(),
                name: ts_simple_ast_extra_1.getName(arg)
            };
        }
    }).filter(a => !!a);
    // missingArgs contains  metadata regarding those nodes that must implicitly move
    const missingArgs = args.map((arg, index) => {
        if (index >= reorder.length) {
            const newIndex = findNextFreeIndexFor(index);
            return {
                index: newIndex,
                originalIndex: index,
                arg: arg.getText(),
                name: ts_simple_ast_extra_1.getName(arg)
            };
        }
    }).filter(a => !!a);
    reorderedArgs = reorderedArgs.concat(missingArgs);
    const replacements = [];
    args.forEach((a, index) => {
        const arg = reorderedArgs.find(r => r.index === index);
        if (arg) {
            replacements.push({ node: args[arg.index], text: arg.arg });
        }
        else {
            log('changeCallArgs ignoring arg: ' + a.getKindName() + ' - text: ' + a.getText() + ' - arg found: ' + (arg && arg.remove));
        }
    });
    for (const r of replacements) {
        r.node.replaceWithText(r.text);
    }
}
//# sourceMappingURL=reorderParams.js.map